//-----------------------------------------------------------------------------
// Torque Game Engine 
// Copyright (C) GarageGames.com, Inc.
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------
// Misc. server commands avialable to clients
//-----------------------------------------------------------------------------

//-----------------------------------------------------------------------------

function serverCmdToggleCamera(%client)
{
   %control = %client.getControlObject();
   if (%control == %client.player)
   {
      %control = %client.camera;
      %control.mode = toggleCameraFly;
   }
   else
   {
      %control = %client.player;
      %control.mode = observerFly;
   }
   %client.setControlObject(%control);
}

function serverCmdDropPlayerAtCamera(%client)
{
   if ($Server::TestCheats || isObject(EditorGui))
   {
      %client.player.setTransform(%client.camera.getTransform());
      %client.player.setVelocity("0 0 0");
      %client.setControlObject(%client.player);
   }
}

function serverCmdDropCameraAtPlayer(%client)
{
   %client.camera.setTransform(%client.player.getEyeTransform());
   %client.camera.setVelocity("0 0 0");
   %client.setControlObject(%client.camera);
}


//-----------------------------------------------------------------------------

function serverCmdSuicide(%client)
{
   //--------------------------------------
   // COGMAC
   //--------------------------------------

   //if (isObject(%client.player))
   //   %client.player.kill("Suicide");

   //--------------------------------------
   // COGMAC
   //--------------------------------------
}   

function serverCmdPlayCel(%client,%anim)
{
   if (isObject(%client.player))
      %client.player.playCelAnimation(%anim);
}

function serverCmdPlayDeath(%client)
{
   if (isObject(%client.player))
      %client.player.playDeathAnimation();
}

//--------------------------------------------------------------------------
// COGMAC
//--------------------------------------------------------------------------

function serverCmdHighlight(%client)
{
   %player = %client.player;

   %mouseVec = %player.getEyeVector();
   %cameraPoint = %player.getEyeTransform();

   // Search for anything that is selectable.
   %searchMasks = $TypeMasks::ItemObjectType | $TypeMasks::StaticShapeObjectType | $TypeMasks::PlayerObjectType;

   // Scan for a target.
   %scanTarg = ScanForTarget( %client, %mouseVec, %cameraPoint, %searchMasks );

   // a target in range was found so select it
   if (%scanTarg)
   {
      %targetObject = firstWord(%scanTarg);
      if( %targetObject != %client.getSelectedObject() )
      {
         ///COGLOG_EVENT( %client.player, "SELECT", %targetObject );
      }
      
      %client.setSelectedObject(%targetObject);
      %client.setSelectedZ( getWord( %scanTarg, 3 ) );
   }
   else {
      ///if( %client.getSelectedObject() != 0 ) COGLOG_EVENT( %client.player, "SELECT", "NULL" );
      %client.setSelectedObject( 0 );
   }
}

function ScanForTarget( %client, %mouseVec, %cameraPoint, %searchMasks )
{
   //Determine how far should the picking ray extend into the world?
   %selectRange = 2000;   

   // scale mouseVec to the range the player is able to select with mouse
   %mouseScaled = VectorScale(%mouseVec, %selectRange);

   // cameraPoint = the world position of the camera
   // rangeEnd = camera point + length of selectable range
   %rangeEnd = VectorAdd(%cameraPoint, %mouseScaled);

   // Search for objects within the range that fit the masks above. If we are
   // in first person mode, we make sure player is not selectable by setting
   // fourth parameter (exempt  from collisions) when calling ContainerRayCast
   %player = %client.player;   
   if (%client.isFirstPerson())
   {
      %scanTarg = ContainerRayCast (%cameraPoint, %rangeEnd, %searchMasks, %player);
   }
   else //3rd person - player is selectable in this case
   {
      %scanTarg = ContainerRayCast (%cameraPoint, %rangeEnd, %searchMasks);
   }

   return %scanTarg;
}

function serverCmdSelectObjectFirstPerson(%client)
{
   %player = %client.player;
   %player.clearProximityCommand();

   %mouseVec = %player.getEyeVector();
   %cameraPoint = %player.getEyeTransform();

   serverCmdSelectObject(%client, %mousePos, %mouseVec, %cameraPoint);
}

// object selection additions
function serverCmdSelectObject(%client, %mousePos, %mouseVec, %cameraPoint)
{
   // Clear a pending proximity command.
   %player = %client.player;
   %player.clearProximityCommand();
   %player.stop();

   // Search for anything that is selectable.
   %searchMasks = $TypeMasks::ItemObjectType | $TypeMasks::StaticShapeObjectType | $TypeMasks::PlayerObjectType;

   // Scan for a target.
   %scanTarg = ScanForTarget( %client, %mouseVec, %cameraPoint, %searchMasks );

   // a target in range was found so select it
   if (%scanTarg)
   {
      %targetObject = firstWord(%scanTarg);
      echo("SELECTED:" SPC %client.player.getName() SPC %targetObject.getName() SPC %scanTarg);
      %client.setSelectedObject(%targetObject);
      %client.player.lastClick = getWords( %scanTarg, 1, 3 );

      // We have a PutDown pending.
      if( %client.player.pendingDown != 0 )
      {
         // Cancel putdown if clicks again on pending object.
         if( %targetObject == %client.player.pendingDown )
         {
            ClearPendingAction(%client);
         }
         else {
            CompletePutDown(%client);
            return;
         }
      }

      // We have a Give pending.
      if( %client.player.pendingGive != 0 )
      {
         // Cancel give if clicks again on pending object.
         if( %targetObject == %client.player.pendingGive )
         {
            ClearPendingAction(%client);
         }
         else {
            CompleteGive(%client);
            return;
         }
      }

      // Toggle the cursor on.
      CommandToClient( %client, 'EnableMouseClicks' );

      // Show GUI.
      CommandToClient( %client, 'ShowInteractionGui', %mousePos );
      return;
   }

   // Cancel when clicking outside of the popup.
   CommandToClient( %client, 'CleanupPopup' );

   // Cancel anything pending when not clicking on an object.
   ClearPendingAction(%client);
}

function ClearPendingAction(%client)
{
   %client.player.pendingDown = 0;
   %client.player.pendingGive = 0;

   // Return to the default cursor.
   CommandToClient( %client, 'SetCrosshair', "default" );
}

function serverCmdMovePlayerStart(%client)
{
   // Clear a pending proximity command.
   %player = %client.player;
   %player.clearProximityCommand();
   %player.stop();

   // Get up from sitting on something.
   if( %player.getDataBlock().isSitting )
   {
      // Do NOT get up if the player is selecting something in hand.
      // Often players accidentally click right instead of left!
      %selected = %client.getSelectedObject();
      if( %selected.isMounted() && ( %selected.getObjectMount() == %client.player ) )
      {
         return;
      }

      // Do NOT get up if player is facing something right in front of the chair.
      // Often players accidentally click right instead of left!
      %chair = %player.getObjectMount();
      %forward = %chair.getForward();
      %forward = VectorScale(%forward, 2);
      %startPoint = %chair.getPosition();
      %endPoint = VectorAdd(%startPoint, %forward);
      %searchMasks = $TypeMasks::StaticObjectType;
      %hit = ContainerRayCast (%startPoint, %endPoint, %searchMasks, %chair);
      if(%hit)
      {
         %obj = getWord(%hit, 0);
         if( %player.isInFOV( %obj, 90 ) ) 
         {
            return;
         }
      }

      %player.getDataBlock().isSitting = false;
	  %vec = %player.getEyeVector();
	  %vec = VectorScale(%vec, 100);
	  %dest = VectorAdd( %player.getPosition(), %vec );

	  // If we were mounted, dismount.
	  %obj = %player.getObjectMount();
	  %obj.unmountObject( %player );

	  COGLOG_ACTION( %client.player, "GETOFF", %obj );
	  COGLOG_STATECHANGE( %client.player, "SITTINGON", "NULL" );

	  // Move to the getting up node.
	  if( %obj.slotExists(1) ) %xform = %obj.getSlotTransform(1);
	  else %xform = %player.getDataBlock().lastStandingXform;
	  %player.setTransform( %xform );
	  COGLOG_STATECHANGE( %client.player, "ATPOS", COGLOG_POS(%player.getPosition()) );

	  %player.FacePosition( %dest );
   }

   COGLOG_ACTION( %client.player, "GOTO", "NULL" );
   %client.player.isMoving = true;
   ReportPosition( %client.player );

   // Cancel popup when moving.
   CommandToClient( %client, 'CleanupPopup' );
}

function ReportPosition( %player )
{
   COGLOG_STATECHANGE( %player, "ATPOS", COGLOG_POS(%player.getPosition()) );
   if( %player.isMoving ) schedule( 100, 0, ReportPosition, %player );
}

function serverCmdMovePlayerEnd(%client)
{
   COGLOG_ACTION( %client.player, "STOP", "NULL" );
   COGLOG_STATECHANGE( %client.player, "ATPOS", COGLOG_POS(%client.player.getPosition()) );
   %client.player.isMoving = false;
}

function serverCmdCleanupPopup(%client)
{
   CommandToClient( %client, 'CleanupPopup' );
}

function MoveTo( %player, %pos, %obj )
{
   // Get up from sitting on something.
   if( %player.getDataBlock().isSitting )
   {
      %player.getDataBlock().isSitting = false;

	  // If we were mounted, dismount.
	  %obj = %player.getObjectMount();
	  %obj.unmountObject( %player );

	  COGLOG_ACTION( %player, "GETOFF", %obj );
	  COGLOG_STATECHANGE( %player, "SITTINGON", "NULL" );

	  // Move to the getting up node.
	  if( %obj.slotExists(1) ) %xform = %obj.getSlotTransform(1);
	  else %xform = %player.getDataBlock().lastStandingXform;
	  %player.setTransform( %xform );
	  COGLOG_STATECHANGE( %player, "ATPOS", COGLOG_POS(%player.getPosition()) );

	  %player.FacePosition( %pos );
   }

   COGLOG_STATECHANGE( %player, "ATPOS", COGLOG_POS(%player.getPosition()) );
   COGLOG_STATECHANGE( %player, "DEST", COGLOG_POS(%pos) );
   COGLOG_ACTION( %player, "GOTO", %obj );

   %player.setMoveDestination( %pos, false );
}

function serverCmdOnPickup(%client)
{
   %targetObject = %client.getSelectedObject();
   %player = %client.player;

   // Don't pick up something you're already holding!
   %mounted = %player.getMountedObject( 0 );
   if( %mounted == %targetObject ) 
   {
      CommandToClient( %client, 'CleanupPopup' );
      return;
   }

   // Move closer and then pick it up.
   if( IsOutOfReach( %player, %targetObject ) )
   {
      %cmd = %targetObject @ ".getDataBlock().OnPickup( " @ %client @ ", " @ %player @ ", " @ %targetObject @ ");";
      %player.setProximityCommand( %targetObject.getPosition(), 0.95 * GetReachDist( %player, %targetObject ), %cmd );
      MoveTo( %player, %targetObject.getPosition(), %targetObject );
      CommandToClient( %client, 'CleanupPopup' );
   }

   // Close enough to pick it up right now.
   else %targetObject.getDataBlock().OnPickup( %client, %player, %targetObject );
}

function serverCmdOnPutDown(%client)
{
   %targetObject = %client.getSelectedObject();
   %player = %client.player;

   // Must be holding it to put it down!
   %mounted = %player.getMountedObject( 0 );
   if( %mounted != %targetObject ) 
   {
      %player.pendingDown = %targetObject;
      CompletePutDown(%client);
      return;
   }

   // Remember what we want to put down, and
   // wait for user to specify where to put it.
   %player.pendingDown = %targetObject;
   CommandToClient( %client, 'CleanupPopup' );

   // Change the cursor.
   CommandToClient( %client, 'SetCrosshair', "down" );
}

function CompletePutDown(%client)
{
   // Nothing to put down.
   if( %client.player.pendingDown == 0 ) return;

   // Clear pending put down.
   ClearPendingAction(%client);

   // Put something down that we are carrying.
   %player = %client.player;
   %targetObject = %client.getSelectedObject();
   %mounted = %player.getMountedObject( 0 );
   if( %mounted > 0 )
   {
      // Don't put something down on top of itself!
      if( %mounted == %targetObject )
      {
         CommandToClient( %client, 'CleanupPopup' );
         return;
      }

      // Move closer and then put it down.
      if( IsOutOfReach( %player, %targetObject ) )
      {
         %cmd = %mounted @ ".getDataBlock().OnPutDown( " @ %client @ ", " @ %player @ ", " @ %mounted @ ", " @ %targetObject @ ", \"" @ %client.player.lastClick @ "\", null );";
         %player.setProximityCommand( %targetObject.getPosition(), 0.95 * GetReachDist( %player, %targetObject ), %cmd );
         MoveTo( %player, %targetObject.getPosition(), %targetObject );
         CommandToClient( %client, 'CleanupPopup' );
      }

      // Close enough to put it down right now.
      else 
      {
         %mounted.getDataBlock().OnPutDown( %client, %player, %mounted, %targetObject, %client.player.lastClick, null );
      }
   }
}

function serverCmdOnLookAt(%client)
{
   %player = %client.player;
   %targetObject = %client.getSelectedObject();

   // Move closer and then look at it.
   if( IsOutOfReach( %player, %targetObject ) )
   {
      %cmd = %targetObject @ ".getDataBlock().OnLookAt( " @ %client @ ", " @ %targetObject @ " );";
      %player.setProximityCommand( %targetObject.getPosition(), 0.95 * GetReachDist( %player, %targetObject ), %cmd );
      MoveTo( %player, %targetObject.getPosition(), %targetObject );
      CommandToClient( %client, 'CleanupPopup' );
   }

   // Close enough to look at it right now.
   else %targetObject.getDataBlock().OnLookAt( %client, %targetObject );
}

function serverCmdOnEat(%client)
{
   %player = %client.player;
   %targetObject = %client.getSelectedObject();

   // Move closer and then eat it.
   if( IsOutOfReach( %player, %targetObject ) )
   {
      %cmd = %targetObject @ ".getDataBlock().OnEat( " @ %client @ ", " @ %targetObject @ ");";
      %player.setProximityCommand( %targetObject.getPosition(), 0.95 * GetReachDist( %player, %targetObject ), %cmd );
      MoveTo( %player, %targetObject.getPosition(), %targetObject );
   }

   // Close enough to eat it right now.
   else %targetObject.getDataBlock().OnEat( %client, %targetObject );
   CommandToClient( %client, 'CleanupPopup' );
}

function serverCmdOnGive(%client)
{
   %targetObject = %client.getSelectedObject();
   %player = %client.player;

   // Must be holding it to give it!
   %mounted = %player.getMountedObject( 0 );
   if( %mounted != %targetObject ) 
   {
      %player.pendingGive = %targetObject;
      CompleteGive(%client);
      return;
   }

   // Remember what we want to put down, and
   // wait for user to specify where to put it.
   %player.pendingGive = %targetObject;
   CommandToClient( %client, 'CleanupPopup' );

   // Change the cursor.
   CommandToClient( %client, 'SetCrosshair', "give" );
}

function CompleteGive(%client)
{
   // Nothing to give.
   if( %client.player.pendingGive == 0 ) return;

   // Clear pending give.
   ClearPendingAction(%client);

   // Give something that we are carrying.
   %player = %client.player;
   %targetObject = %client.getSelectedObject();
   %mounted = %player.getMountedObject( 0 );
   if( %mounted > 0 )
   {
      // Don't give something to itself!
      if( %mounted == %targetObject )
      {
         CommandToClient( %client, 'CleanupPopup' );
         return;
      }

      // Move closer and then give it.
      if( IsOutOfReach( %player, %targetObject ) )
      {
         %cmd = %targetObject @ ".getDataBlock().OnGive( " @ %client @ ", " @ %player @ ", " @ %mounted @ ", " @ %targetObject @ ", \"" @ %client.player.lastClick @ "\" );";
         %player.setProximityCommand( %targetObject.getPosition(), 0.95 * GetReachDist( %player, %targetObject ), %cmd );
         MoveTo( %player, %targetObject.getPosition(), %targetObject );
      }

      // Close enough to give it right now.
      else %targetObject.getDataBlock().OnGive( %client, %player, %mounted, %targetObject, %client.player.lastClick );
   }

   else CommandToClient( %client, 'CleanupPopup' );
}

function serverCmdOnTouch(%client)
{
   %player = %client.player;
   %targetObject = %client.getSelectedObject();

   // Move closer and then touch it.
   if( IsOutOfReach( %player, %targetObject ) )
   {
      %cmd = %targetObject @ ".getDataBlock().OnTouch( " @ %client @ ", " @ %player @ ", " @ %targetObject @ ");";
      %player.setProximityCommand( %targetObject.getPosition(), 0.95 * GetReachDist( %player, %targetObject ), %cmd );
      MoveTo( %player, %targetObject.getPosition(), %targetObject );
      CommandToClient( %client, 'CleanupPopup' );
   }

   // Close enough to touch it right now.
   else %targetObject.getDataBlock().OnTouch( %client, %player, %targetObject );
}

function serverCmdOnSit(%client)
{
   %player = %client.player;
   %targetObject = %client.getSelectedObject();

   // Don't sit on something you're holding!
   %mounted = %player.getMountedObject( 0 );
   if( %mounted == %targetObject ) 
   {
      CommandToClient( %client, 'CleanupPopup' );
      return;
   }

   // Move closer and then sit on it.
   if( IsOutOfReach( %player, %targetObject ) )
   {
      %cmd = %targetObject @ ".getDataBlock().OnSit( " @ %client @ ", " @ %player @ ", " @ %targetObject @ ");";
      %player.setProximityCommand( %targetObject.getPosition(), 0.95 * GetReachDist( %player, %targetObject ), %cmd );
      MoveTo( %player, %targetObject.getPosition(), %targetObject );
   }

   // Close enough to sit on it right now.
   else %targetObject.getDataBlock().OnSit( %client, %player, %targetObject );
   CommandToClient( %client, 'CleanupPopup' );
}

function serverCmdOnGoto(%client)
{
   %player = %client.player;
   %targetObject = %client.getSelectedObject();

   // Don't goto something you're holding!
   %mounted = %player.getMountedObject( 0 );
   if( %mounted == %targetObject ) 
   {
      CommandToClient( %client, 'CleanupPopup' );
      return;
   }

   MoveTo( %player, %targetObject.getPosition(), %targetObject );
   CommandToClient( %client, 'CleanupPopup' );
}

function serverCmdPlay(%client, %anim)
{
   %client.player.setActionThread( %anim, false, false );
}

function serverCmdClear(%client)
{
   %client.player.clearActionThread();
}

function IsOutOfReach( %player, %obj )
{
   %playerPos = getWords( %player.getPosition(), 0, 1 ) SPC "0";
   %objPos = getWords( %obj.getPosition(), 0, 1 ) SPC "0";

   %mult = 1.0;
   if( %player.getDataBlock().isSitting ) %mult = 1.25;

   %dist = VectorDist( %playerPos, %objPos );
   if( %dist > (GetReachDist( %player, %obj ) * %mult) ) return true;
   return false;
}

function GetReachDist( %player, %obj )
{
echo( "REACH:" SPC %player.getName() SPC %obj.getName() SPC %obj.getDataBlock().reachDist );
   return %obj.getDataBlock().reachDist + %obj.getRadius();
}

function ProcessChatMessage(%client, %text )
{
   // Determine who the speaker is facing.
   for( %i=0; %i < $listenerList.nObjs; %i++ )
   {
      // Speaker cannot face himself!
      %obj = $listenerList.lstObjs[%i];
      if( %obj == %client.player ) continue;

      // Object is outside of the speaker's FOV.
      if( !%client.player.isInFOV( %obj, 180 ) ) continue;

      // Speaker is facing the object.
      COGLOG_EVENT( %client.player, "FACING", %obj );
   }

   COGLOG_STATECHANGE( %client.player, "FORWARDDIR", COGLOG_POS(%client.player.getEyeVector()) );
   COGLOG_STATECHANGE( %client.player, "ATPOS", COGLOG_POS(%client.player.getPosition()) );
   COGLOG_SPEECHACT( %client.player, %text );

   echo( "[" @ detag(%client.nameBase) @ "]" SPC %text );

   // Process food orders.
   %counterTrigger = nameToId("counterTrigger");
   if( %counterTrigger.containsObject( %client.player ) )
   {
      %foodTrigger = nameToId("foodTrigger");
      %counterTrigger.getDataBlock().ProcessOrder( %counterTrigger, %foodTrigger, %text, "food", %client.player, nameToId("chef") );
   }

   // Process drink orders.
   %barTrigger = nameToId("barTrigger");
   if( %barTrigger.containsObject( %client.player ) )
   {
      %drinkTrigger = nameToId("drinkTrigger");
      %barTrigger.getDataBlock().ProcessOrder( %barTrigger, %drinkTrigger, %text, "drink", %client.player, nameToId("bartender") );
   }
}

function PlaySound( %file, %pos )
{
   %sound = new AudioEmitter() 
   {
      position = %pos;
      fileName = %file;
      description = AudioClose3d;  //for loud sound
   };
   schedule(3000,0,deletethis, %sound); //delete the sound in 3 sec.
}

function deletethis(%obj)
{
   %obj.delete();
}

function serverCmdEnableMouseClicks(%client)
{
   // Toggle the cursor on.
   CommandToClient( %client, 'EnableMouseClicks' );
}

function serverCmdPlayerExit(%client)
{
   CommandToClient( %client, 'PlayerExit' );
}

function serverCmdGetServerName(%client)
{
   CommandToClient( %client, 'ShowServerName', $Server::DedicatedName );
}

function SetCurrentBalance(%client, %bal)
{
   %client.money = %bal;
   %client.player.SetPlayerAccountValue( "money", %bal );
   CommandToClient( %client, 'DisplayCurrentBalance', %bal );

   COGLOG_STATECHANGE( %client.player, "BALANCE", %bal );
}

function serverCmdLogAdjectives(%client, %other, %adj1, %adj2, %adj3)
{
   COGLOG_DESCRIPTION( %client.player, %other, %adj1 );
   COGLOG_DESCRIPTION( %client.player, %other, %adj2 );
   COGLOG_DESCRIPTION( %client.player, %other, %adj3 );
}

function serverCmdSaveProfile(%client, %other, %report)
{
   // Log survey submissions.
   COGLOG_EVENT( %client.player, "SUBMITTED_SURVEY", "NULL" );
   COGLOG_LOCALTIME();

   COGLOG_INFO( %client.player, "SubmittedSurveyResults", %report );

   // Find the referrent.
   %player = null;
   %count = ClientGroup.getCount();	
   for ( %i = 0; %i < %count; %i++ )
   {
      %curClient = ClientGroup.getObject( %i );
      if( %curClient.player.getName() $= %other )
      {
         %otherClient = %curClient;
         %email = %curClient.email;
         %player = %curClient.player;
         break;
      }
   }

   // Player not found.
   if( %player == null ) return;

   // Send report to client.
   CommandToClient( %otherClient, 'DisplayProfile', %report, %email );
}

function serverCmdLogLocalTime(%client)
{
   COGLOG_LOCALTIME();
}

function serverCmdLogInfo(%client, %tag, %info)
{
   COGLOG_INFO( %client.player, %tag, %info );
}

function serverCmdNotifyPlayerArriving(%client, %role)
{
   %client.joinStatus = "loading";
   %client.joinRole = %role;
   SetNarration( %client, "[" @ %role SPC "is loading the game and reading objectives...]" );
}

function serverCmdNotifyPlayerJoin(%client)
{
   %client.joinStatus = "joined";
   SetNarration( %client, "[" @ %client.player.getName() SPC "has joined the game]" );

   // Send status of other clients.   
   %count = ClientGroup.getCount();	
   for ( %i = 0; %i < %count; %i++ )
   {
      %curClient = ClientGroup.getObject( %i );
      if( %curClient.joinStatus $= "loading" )
      {
         serverCmdNotifyPlayerArriving(%curClient, %curClient.joinRole);
      }
   }
}

function serverCmdWhoElse(%client)
{
   // Find the other client.
   %count = ClientGroup.getCount();	
   for ( %i = 0; %i < %count; %i++ )
   {
      %curClient = ClientGroup.getObject( %i );
      if( %curClient != %client )
      {
         ///CommandToClient( %client, 'RespondWhoElse', %curClient.email, %curClient.credit );
         return;
      }
   }

   // Did not find anyone else.
}

function serverCmdNotifyStartedSurvey(%client)
{
   SetNarration( %client, "[When you are ready, go to the Objective Complete square outside.]" );

   // Keep track of how many games this player has played.
   %numGames = %client.player.GetPlayerAccountValue( "games" );
   %client.player.SetPlayerAccountValue( "games", (%numGames + 1) );

   // Now you get your money!
   if( $Server::AccountNetworkUp )
   {
      %client.player.UpdatePlayerAccount( %client.email );
   }
}

//--------------------------------------------------------------------------
// COGMAC
//--------------------------------------------------------------------------













